// Propeller Display with H8S/2633 Microcontroller-Board
// Software for the terminal
// Entry H3210
// Renesas H8 Design Contest 2003
//
// Johann Gysin

#include "ioh82633.h"
#include "lcd.h"
#include "sci.h"

#define enable_irqs()  asm("andc #0x3F,ccr") // Clear I-Bit in
#define disable_irqs() asm("orc #0xC0,ccr")  // Set I-Bit in CCR

// Global variables
const char Characters[] = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 228, 246, 252, ' ', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 196, 214, 220, ' ', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ' ', '.', ',', '?', '!', ':', ';', '-', '+', '#', '*', '(', ')', 96, '"', '_', '@', '&', '$', '%', '/', '<', '>', '=', ' ', 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, ' ', 127, ' '};
const char CharactersLength = 109;
const char StartString[] = "StarT"; // RF: Startsequency
const char StartString_Length = 5;  // RF: Length of startsequency
char RFText[400];                   // RF: Per radio (loop module) received string
const int RFText_MaxLength = 400;   // RF: Max length of string
unsigned int timer_tick = 0;        // Counter for timerinterrupts (0.5ms)
unsigned int seconds = 0;
unsigned int seconds_old = -1;
unsigned char dummy;
unsigned char line;
unsigned char character;
int RotationSwitch_counter = 0;     // Counter for Rotation-Switch position
unsigned int timer_sec = 0;         // Second-Timer
unsigned char SwitchStatus;         // Actual switch status
unsigned char SwitchStatus_Before;  // Last switch status
unsigned char UserInput = 0;        // Usertext to transmit to the propeller display
int Switch_Shift_Counter = 0;       // Digital Filter for Shift-Switch
int Switch_Delete_Counter = 0;      // Digital Filter for Delete-Switch
char DeleteLock = 0;                // Digital Filter for DeleteLock-Switch
char ShiftLock = 0;                 // Digital Filter for ShiftLock-Switch
unsigned char SCI2_String[400];     // From SCI2 received string
int SCI2_String_Length = 0;         // Character count in the string
const int SCI2_String_LengthMax = 400;

// Functions
void Timer_Tick(void);
void Laufschrift_Update(void);
void RotationSwitch_Read(void);
void ReadSwitch(void);
void AddChar(unsigned char *MainString, unsigned char AddChar); // Adds a character to the string
void RemoveLastChar(unsigned char *MainString);                 // Removes last character from the string
void RF_SendStr(unsigned char *str);                            // Sends a string to the air with the loop module
void WaitTime(unsigned int Time);                               // Waits x*millisecond
void UserText_Update(unsigned char *str);                       // If the user entered a character
void StringInit(unsigned char *str, unsigned int str_length);   // Fills a string with null characters

// Interrupt-Functions

#pragma interrupt
void Timer_Tick(void)
  {
  timer_tick++;
  
  // 2ms 
  if ((timer_tick % 4) == 0)
    {
    RotationSwitch_Read();
    }
  
  // 10ms    
  if ((timer_tick % 20) == 0) 
    {
    ReadSwitch();
    }
  
  // 1000ms    
  if ((timer_tick % 2000) == 0)
    {
    timer_tick = 1;
    timer_sec++;
    }
    
  dummy = TCSR0;
  TCSR0 = TCSR0 & 191; // Clear Compare Match Flag    
  }

int main (void)
  {
  int index;
  int RotationSwitchTemp;
  int AlphabetStart = 0;
  unsigned char UserText[20];
  const unsigned int UserText_Length = 20; 
  int CursorPosition = 0;
  unsigned char character = 0;
  int length = 0;
    
  line = 1;
  character = 1;
  
  // PWM for LCD-Contrast
  MSTPCRB = 249;           // Start PWM-Counter (PWM register are writeable only when PWM was started)
  DACR0 = 60;              // Activate PWM-Ourputs & Invert Modus
  DACNT0 = 2;              // Grant access to DADRA (register has a double function)
  DADRA0 = 55000 + 3;      // Pulse length (Hi-Byte) + CFS = 1 and NA = 1
  DADRB0 = 1;              // Grant access to DACR and DACNT (register has a double function)
  
  // SCI init
  SCI0_Init();
  SCI2_Init();
  
  // Display init
  LCD_Init(1);
  LCD_Control(1, 1, 0, 0); // Display On, Cursor Off, Cursor Blink Off
    
  // Counter Init
  MSTPCRA = MSTPCRA & 239; // Clear Bit 4 to start timer (TMR0, TMR1)
  TCORA0 = 144;            // 1/(Phi/64)*144 = 500us (Timer for Timer-Tick)
  TCR0 = 64 | 8 | 2;       // Enable CMFA Interrupt, Clear @ Compare Match A, Internal Clock Phi/64
 
  // Show Splash-Screen
  LCD_WriteStr(1, 1, " Propeller Display  ");
  LCD_WriteStr(2, 1, " Renesas Design     ");
  LCD_WriteStr(3, 1, " Contest 2003  H3210");
  LCD_WriteStr(4, 1, " J. Gysin           ");
  
  // Give out Splash-Screen to serial interface
  SCI2_WriteLn("");
  SCI2_WriteLn("Propeller Display");
  SCI2_WriteLn("");
  SCI2_WriteLn("Renesas Design Contest 2003");
  SCI2_WriteLn("Entry H3210 by J. Gysin");
  
  WaitTime(2000);          // Wait 2 seconds
  
  // Activate Switches
  PEDDR = 0;               // Port_E IO's as outputs
  PEPCR = PEPCR | 127;     // Activate Pull-Up Resistors
  
  // Loop Module (Radio)
  PGDDR = 8 | 4 | 2 | 1; // Port_G IO's as outputs (Bit 0, 1, 2, 3)
  PGDR = 0;              // Receive Mode, Channel 0, PWR_Down
  
  enable_irqs();         // Enable Interrupts
  
  // UserText Array init
  StringInit(UserText, UserText_Length);
  StringInit(SCI2_String, SCI2_String_LengthMax);
  
  // Display; Clear Splash-Screen
  LCD_WriteStr(1, 1, "                    ");
  LCD_WriteStr(2, 1, "                    ");
  LCD_WriteStr(3, 1, "                    ");
  LCD_WriteStr(4, 1, "                    ");
  LCD_WriteChar(LCD_SetCursor(1, 1), 127); // Show Cursor
  LCD_Control(1, 1, 1, 1); // Display On, Cursor On, Cursor Blink On
    
  while (1)
    {
    // If a string from the serial interface was received
    length = SCI2_GetRecBufferCount();
    if (length > 0)
      {
      for (index=0; index<length; index++)
        {
        character = SCI2_GetRecBufferChar();
        if ((character != 13) & (character != 10))
          {
          if (character != 8) // If character was not a delete...
            {
            if (SCI2_String_Length < SCI2_String_LengthMax)
              {
              AddChar(SCI2_String, character);
              SCI2_String_Length++;
              }
            }
          else
            {
            RemoveLastChar(SCI2_String);
            SCI2_String_Length--;
            }
          }
        else
          {
          if (character == 13)
            {
            RF_SendStr(SCI2_String);
            SCI2_WriteLn("Sent Text!");
            }
          StringInit(SCI2_String, SCI2_String_LengthMax);
          SCI2_String_Length = 0;
          }
        }
      }
      
    // If the user rotated or pushed the rotation switch...    
    if (UserInput)
      {
      
      // Read Cursorposition
      CursorPosition = RotationSwitch_counter;
        
      // Upper Cursorlimit
      if (CursorPosition > (CharactersLength-1))
        {
        CursorPosition = 0;
        }
      
      // Lower Cursorlimit
      if (CursorPosition < 0)
        {
        CursorPosition = CharactersLength-1;
        AlphabetStart = CharactersLength - 17;
        }
      
      // Show Alphabet
      if ((CursorPosition - AlphabetStart) > 16)
        {
        AlphabetStart = CursorPosition - 16;
        }
      
      if ((CursorPosition - AlphabetStart) < 3)
        {
        AlphabetStart = CursorPosition - 3;
        if (AlphabetStart < 0)
          {
          AlphabetStart = 0;
          }
        }
        
      for (index=0; index<20; index++)
        {
        LCD_SetCursor(4, (index+1));
        if ((AlphabetStart+index)<CharactersLength)
          {
          LCD_WriteChar(4, Characters[(AlphabetStart+index)]); // Show a window of the alphabet on the LCD
          }
        else
          {
          LCD_WriteChar(4, ' '); // If end of array is reached, fill rest of LCD-Line with spaces
          }
        }
        
      LCD_SetCursor(4, (CursorPosition - AlphabetStart)+1);
        
      RotationSwitch_counter = CursorPosition;
        
      // Push on the rotation switch
      if ((SwitchStatus & 1) == 0)
        {
        if (Characters[CursorPosition] != 127)
          {  
          AddChar(UserText, Characters[CursorPosition]);
          UserText_Update(UserText);
          LCD_SetCursor(4, (CursorPosition - AlphabetStart)+1);
          }
        else
          {
          RF_SendStr(UserText);
          LCD_Control(1, 1, 0, 0);                    // Display On, Cursor Off, Cursor Blink Off
          LCD_WriteStr(1, 1, "                    ");
          LCD_WriteStr(2, 1, "   Sent Text!       ");
          LCD_WriteStr(3, 1, "                    ");
          WaitTime(1000);
          LCD_WriteStr(2, 1, "                    ");
          UserText_Update(UserText);
          LCD_SetCursor(4, (CursorPosition - AlphabetStart)+1); // Set cursor on alphanet
          LCD_Control(1, 1, 1, 1);                    // Display On, Cursor On, Cursor Blink On
          }
        }
        
      UserInput = 0;
      }
      
      // Delete Switch
      if (Switch_Delete_Counter > 10)
        {
        if (DeleteLock == 0)
          {
          RemoveLastChar(UserText);
          LCD_WriteStr(1, 1, "                    ");
          LCD_WriteStr(2, 1, "                    ");
          LCD_WriteStr(3, 1, "                    ");
          LCD_WriteChar(LCD_SetCursor(1, 1), 127);    // Show Cursor
          UserText_Update(UserText);                  // Show String
          DeleteLock = 1;
          }
        }
        
      // Delete whole string is delete was pressed for longer than 2 seconds
      if (Switch_Delete_Counter > 200)
        {
        StringInit(UserText, UserText_Length);
        LCD_WriteStr(1, 1, "                    ");
        LCD_WriteStr(2, 1, "                    ");
        LCD_WriteStr(3, 1, "                    ");
        LCD_WriteChar(LCD_SetCursor(1, 1), 127);    // Show Cursor
        UserText_Update(UserText);                  // Show String
        Switch_Delete_Counter = 0;
        DeleteLock = 0;
        }  
        
      // Shift Switch
      if (Switch_Shift_Counter > 10)
        {
        if (ShiftLock == 0)
          {
          RotationSwitchTemp = 0;
          
          // Jump to upper characters in the alphabet
          if (RotationSwitch_counter < 30)                              // If lower character...
            {
            RotationSwitchTemp = 30;
            }
        
          // Jump to lower characetrs in the alphabet
          if ((RotationSwitch_counter > 29) && (RotationSwitch_counter < 60)) // If upper character and no special characetr...
            {
            RotationSwitchTemp = -30;
            }
          
          // Execute jump
          RotationSwitch_counter = RotationSwitch_counter + RotationSwitchTemp;
          
          ShiftLock = 1;
          }
        }
        
      SwitchStatus_Before = SwitchStatus; // Store current status
    
    }
  }
  
void RotationSwitch_Read(void)
  {
  unsigned char SwitchState;
  static unsigned int SwitchState_before = 0;
  
  SwitchState = PORTE & 6;                 // Take bits 1 and 2 from port E
  SwitchState = SwitchState >> 1;
  
  if (SwitchState != SwitchState_before)   // If state changed
    {      
    switch (SwitchState)
      {
      // On a step
      case 1: if (SwitchState_before == 3) // If switch was idle (11)
                {
                RotationSwitch_counter++;  // One step to the right
                UserInput = 1;             // Signalisation for main-program
                }
              else
                {
                RotationSwitch_counter--;  // One step to the left
                UserInput = 1;             // Signalisation for main-program
                }
              break;
    
      // On a step
      case 2: if (SwitchState_before == 0) // If switch was idle (00)
                {
                RotationSwitch_counter++;  // One step to the right
                UserInput = 1;             // Signalisation for main-program
                }
              else
                {
                RotationSwitch_counter--;  // One step to the left
                UserInput = 1;             // Signalisation for main-program
                }
              break;
      }
    }
    
  SwitchState_before = SwitchState;
  }
  
void ReadSwitch(void)
  {
  static unsigned char SwitchStatus_Before;
    
  SwitchStatus = PORTE & 127;            // Read Switch Position
      
  if (SwitchStatus != SwitchStatus_Before)
    {
    SwitchStatus_Before = SwitchStatus;
    UserInput = 1;                  // Signalisation for the main program
    }
  
  // Shift  
  if ((SwitchStatus & 8) == 0)
    {
    if (Switch_Shift_Counter < 21)
      {
      Switch_Shift_Counter++;
      }
    }
  else
    {
    if (Switch_Shift_Counter > 0)
      {
      Switch_Shift_Counter = 0;
      }
    else
      {
      ShiftLock = 0;
      }
    }
    
  // Delete  
  if ((SwitchStatus & 16) == 0)
    {
    if (Switch_Delete_Counter < 210)
      {
      Switch_Delete_Counter++;
      }
    }
  else
    {
    if (Switch_Delete_Counter > 0)
      {
      Switch_Delete_Counter = 0;
      }
    else
      {
      DeleteLock = 0;
      }
    }
    
  }
  
void AddChar(unsigned char *MainString, unsigned char AddChar)
  {
  if (*MainString != '\0')       // If first character isn't null...
    {  
    do
      {
      MainString++;
      }
    while (*MainString != '\0'); // ...until null is reached
    }
    
  *MainString = AddChar;         // Put character on the previous position of the null
  MainString++;                  // Increment Pointer
  *MainString = '\0';            // Insert null character at the new end of string
  }
  
void RemoveLastChar(unsigned char *MainString)
  {
  if (*MainString != '\0')       // If first character isn't null...
    {
    do
      {
      MainString++;              // Increment Pointer
      }
    while (*MainString != '\0'); // ...until null is reached
    MainString--;                // Return to last character in the string
    }
    
  *MainString = '\0';            // Insert null character at the new end of string
  }
  
void StringInit(unsigned char *str, unsigned int str_length)
  {
  unsigned int index;
    
  for (index=0; index<str_length; index++)
    {
    str[index] = '\0';
    }
  }
  
void RF_ProcessData(void)
  {
  unsigned char character;
  int index;
  int BufCharacterCount;
  static unsigned char synced = 0;            // 1 => Startsequency was ok
  static unsigned char StartString_Index = 0; // Index in Startsequency
  static unsigned int RF_RecCharacterCount;   // Received character count
    
  BufCharacterCount = SCI0_GetRecBufferCount();
  if (BufCharacterCount > 0)
    {
    for (index=0; index<BufCharacterCount; index++)
      {
      character = SCI0_GetRecBufferChar(); // Get a character from the receive-buffer
      if (synced)
        {
        if (character == 13) // If end of string reached...
          {
          synced = 0;
          }
        else
          {
          if (RF_RecCharacterCount < RFText_MaxLength) // If string isn't full...
            {
            AddChar(RFText, character); // Add character to string
            }
          RF_RecCharacterCount++;
          }
        }
      else
        {
        if (character == StartString[StartString_Index])
          {
          StartString_Index++;
          if (StartString_Index >= StartString_Length) // If startsequency is complete...
            {
            StartString_Index = 0;
            StringInit(RFText, RFText_MaxLength);
            synced = 1;
            RF_RecCharacterCount = 0;
            }
          }
        else
          {
          StartString_Index = 0;
          }
        }  
      }  
    }
  }
  
void RF_SendStr(unsigned char *str)
  {
  unsigned int index;
    
  PGDR = 4 | 1;               // Send Mode, Channel 0, PWR_Up

  WaitTime(5);                // Wait 5ms to start up the loop module
  
  SCI0_WriteChar(170);        // PLL synchronization with 10101010
  SCI0_WriteChar(170);        // PLL synchronization with 10101010
  SCI0_WriteChar(255);        // UART synchronization
  SCI0_WriteStr(StartString); // Send startsequency
  SCI0_WriteStr(str);         // Send userstring
  
  // Send stoppsequency some times
  for (index=0; index<10; index++)
    {
    SCI0_WriteChar(13);   
    SCI0_WriteChar(10); 
    }
  
  WaitTime(100);          // Wait some time
  
  PGDR = 0;               // Receive Mode, Channel 0, PWR_Down
  }
  
void WaitTime(unsigned int Time)
  {
  unsigned int a;
  unsigned int b;

  for (a=0; a<Time; a++)
    {
    for (b=0; b<3650; b++) // Waits about 1ms
     {
     }
    }    
  
  }
  
void UserText_Update(unsigned char *str)
  {
  int LCD_Controller;
  int character = 1;
  int line = 1;
    
  while (*str != '\0')
    {
    LCD_Controller = LCD_SetCursor(line, character);
    LCD_WriteChar(LCD_Controller, *str);
    character++;
    if (character > 20)
      {
      line++;
      character = 1;
      if (line > 3)
        {
        line = 1;
        character = 1;
        }
      }
    str++;
    }
  
  LCD_Controller = LCD_SetCursor(line, character);
  LCD_WriteChar(LCD_Controller, 127);   // Put a cursor (left hand arrow) to the end of the UserString
  }
